#include <sqlextract/commands/command_select.h>
#include <sqlextract/command.h>
#include <sqlextract/filter/filter.h>
#include <sqlextract/util.h>
#include <sqlextract/tables/table.h>
#include <stdio.h>
#include <stdlib.h>

struct command_select_t {
	struct selcollist_t *columns;
	struct filter_t *where;
	struct table_t *table;
	struct select_limit_t *limit;

	void *userdata;
};

struct command_select_iterstorage_t {
	struct command_select_t *csel;
	struct output_formatter_t *formatter;

	long limit;
	long offset;
	int limit_enabled;
};

struct selcollist_t {
	struct filter_t *flt;
	struct selcollist_t *next;
};

struct select_limit_t {
	struct filter_t *limit;
	struct filter_t *offset;
};

static void command_select_destroy(void *cselp);
static int command_select_execute(void *cselp, void *arg,
		struct output_formatter_t *formatter);
static int command_select_output(struct table_row_t *row, void *storage);

struct command_t *command_select_create(struct selcollist_t *columns,
		struct filter_t *where, struct table_t *table,
		struct select_limit_t *limit) {
	struct command_select_t *csel = lf_malloc(sizeof(struct command_select_t));
	csel->columns = columns;
	csel->where = where;
	csel->table = table;
	csel->userdata = NULL;
	csel->limit = limit;

	return command_create(command_select_execute, command_select_destroy,
			(void*) csel);
}
static void command_select_destroy(void *cselp) {
	struct command_select_t *csel = (struct command_select_t *) cselp;
	if (csel == NULL)
		return;
	selcollist_destroy(csel->columns);
	filter_destroy(csel->where);
	select_limit_destroy(csel->limit);
	free(csel);
}
static int command_select_output(struct table_row_t *row, void *storage) {
	struct command_select_iterstorage_t *iterstore =
			(struct command_select_iterstorage_t *) storage;
	struct arbval_t *col_result;
	struct selcollist_t *cur_col;

	if(iterstore->limit_enabled) {
		if(iterstore->limit == 0) {
			return 0;
		}
		if(iterstore->offset > 0) {
			iterstore->offset--;
			return 1;
		}
		iterstore->limit--;
	}

	output_formatter_row(iterstore->formatter);
	for (cur_col = iterstore->csel->columns; cur_col; cur_col = cur_col->next) {
		col_result = filter_process(cur_col->flt, row);
		output_formatter_field(iterstore->formatter, col_result);
	}
	return 1;
}
static int command_select_execute(void *cselp, void *arg,
		struct output_formatter_t *formatter) {
	struct command_select_t *csel = (struct command_select_t *) cselp;
	struct command_select_iterstorage_t iterstore;

	struct arbval_t *tmp_av;

	iterstore.csel = csel;
	iterstore.formatter = formatter;
	iterstore.limit_enabled = 0;

	if(csel->limit != NULL) {
		tmp_av = filter_process(csel->limit->limit, NULL);
		arbval_typecast_integer(tmp_av);
		iterstore.limit = tmp_av->value.integer;

		tmp_av = filter_process(csel->limit->offset, NULL);
		arbval_typecast_integer(tmp_av);
		iterstore.offset = tmp_av->value.integer;

		iterstore.limit_enabled = 1;
	}


	output_formatter_begin(formatter);
	if (csel->table != NULL) {
		table_filter(csel->table, command_select_output, &iterstore,
				csel->where);
	}
	output_formatter_end(formatter);

	return 0;
}

struct selcollist_t *selcollist_create(struct selcollist_t *head,
		struct filter_t *filter) {
	if (head == NULL) {
		head = lf_malloc(sizeof(struct selcollist_t));
		head->flt = filter;
		head->next = NULL;
	} else {
		head->next = selcollist_create(head->next, filter);
	}
	return head;
}
void selcollist_destroy(struct selcollist_t *scol) {
	if (scol == NULL)
		return;
	selcollist_destroy(scol->next);
	filter_destroy(scol->flt);
	free(scol);
}

struct select_limit_t *select_limit_create(struct filter_t *limit,
		struct filter_t *offset) {
	struct select_limit_t *lim = lf_malloc(sizeof(struct select_limit_t));
	lim->limit = limit;
	lim->offset = offset;
	return lim;

}
void select_limit_destroy(struct select_limit_t *lim) {
	if (lim == NULL)
		return;
	filter_destroy(lim->limit);
	filter_destroy(lim->offset);
	free(lim);
}
